;CAML192e.S -- Forth Kernel
; CamelForth for the Zilog eZ80F92
; portions (C) 2003 Douglas Beattie Jr.
;
; ..based on CamelForth for the Zilog Z80
; (C) 1994 Bradford J. Rodriguez
; Permission is granted to freely copy, modify,
; and distribute this program for personal or
; educational use.  Commercial inquiries should
; be directed to the author at 115 First St.,
; #105, Collingwood, Ontario L9Y 4W3, Canada
;
; CAMEL80.S: Code Primitives
;   Source code is for the Zilog Macro Assembler.
;   Forth words are documented as follows:
;x   NAME     stack -- stack    description
;   where x=C for ANS Forth Core words, X for ANS
;   Extensions, Z for internal or private words.
;
; Direct-Threaded Forth model for Zilog Z80
; 16 bit cell, 8 bit char, 8 bit (byte) adrs unit
;        BC =       TOS  'Top Of Stack',holds top Param Stack item
;        HL =       W    working register
;        DE =       IP   Interpreter Pointer
;        SP =       PSP  Param Stack Pointer
;        IX =       RSP  Return Stack Pointer
;        IY =       UP   User area Pointer
;    A, alternate register set = temporaries
;
; Revision history:
;   19 Aug 94 v1.0
;   25 Jan 95 v1.01  now using BDOS function 0Ah
;       for interpreter input; TIB at 82h.
;   02 Mar 95 v1.02  changed ALIGN to ALIGNED in
;       S" (S"); changed ,BRANCH to ,XT in DO.
;   17 NOV 1998 SOURCE CODE CONVERTED TO ZMA v1.2
;       (Zilog Macro Assembler) by D. Beattie Jr.
;   27 FEB 2000 PORTED TO TRS-80 by D. Beattie Jr.
;   25 SEP 2001 PORTED TO eZ80 by D. Beattie Jr.
;   15 NOV 2001 v1.06R is ROMmable eZ80 version.
;   12 MAY 2003 v1.10e for eZ80 Acclaim, ZDS II (ZiLOG) XTools 
;
    ;.CPU    eZ80F92

; frequency must be set properly for baud rate to
; be initialized correctly.
;CPUFREQ EQU 23961600    ;23.9616_MHz
;CPUFREQ EQU  11980800    ;11.9808_MHz
CPUFREQ EQU 20000000	;F92 Development Kit, 20_MHz


    .INCLUDE "eZ80F92.S"    ; CPU Equates

;************************************************************************
    define  Startup,SPACE=ROM, ORG=%E000 ;,org=%0000
    ;define  Onchip,SPACE=RAM ,org=%E000

; ===============================================
; Macros to define Forth headers
; HEAD  label,name,action
; IMMED label,name,action
;    label  = assembler name for this word
;             (special characters not allowed)
;    name   = Forth's name for this word
;    action = code routine for this word, e.g.
;             DOCOLON, or DOCODE for code words
; IMMED defines a header for an IMMEDIATE word.
;
; NOTE: length of name field is now automatically
; calculated by the macro, and is no longer a parameter
;
docode  EQU 0      ; flag to indicate CODE words
link    .SET 0     ; link to previous Forth word

head    MACRO   label,name,action
        DW link
        DB 0
link    .SET $
;;      D B &length,&name
        DB $F-$-1,name
$$:
label:
        IFSAME  action,docode
		ELSE
        call action
        ENDIF
        ENDMACRO head

immed   MACRO   label,name,action
        DW link
        DB 1
link    .SET $
;;      DB length,&name
        DB $F-$-1,name
$$:
label:
        IFSAME action,docode
		ELSE
        call action
        ENDIF
        ENDMACRO immed

; The NEXT macro (7 bytes) assembles the 'next'
; code in-line in every eZ80 CamelForth CODE word.
next    MACRO
        EX DE,HL
		LD DE,(HL)
        EX DE,HL
		INC DE
		INC DE
        JP (HL)
        ENDMACRO next

; NEXTHL is used when the IP is already in HL.
nexthl  MACRO
		LD DE,(HL)
        EX DE,HL
		INC DE
		INC DE
        JP (HL)
        ENDMACRO nexthl


ENDOFMEM EQU 0FFFDH
;ENTRY   EQU  0040h


	SEGMENT Startup; Onchip;Startup

    .ASSUME ADL=0

; Memory map for eZ80 Evaluation Board:
;   0000h       Forth kernel in ROM
;   0E000h      Terminal Input Buffer, 128 bytes (eZ80 on-chip RAM)
;   0E080h      Forth dictionary (eZ80 on-chip RAM)
;   EM-200h     User area, 128 bytes
;   EM-180h     Parameter stack, 128 bytes, grows down
;   EM-100h     HOLD area, 40 bytes, grows down
;   EM-0D8h     PAD buffer, 88 bytes
;   EM-80h      Return stack, 128 bytes, grows down
;   0FFFDh      End of Memory ('EM') (End of RAM)
; See also the definitions of U0, S0, and R0
; in the "system variables & constants" area.
; A task w/o terminal input requires 200h bytes.
; Double all except TIB and PAD for 32-bit CPUs.

; RESET AND INTERRUPT VECTORS ===================

    ;ORG 0000h
    ;JP  ENTRY
;    ORG 08h        ;RST 1 vector
;    RET
;    DW RST08ISR
;    ORG 10h        ;RST 2 vector
;    RET
;    ORG 18h        ;RST 3 vector
;    RET
;           ...etc.  (for later implementation)
;

    ;ORG     ENTRY   ; program ENTRY POINT


ENTRY: ;reset:
;
;   enable eZ80 On-Chip RAM!!
;
    LD  A,80h
    OUT0 (RAM_CTL),A


; ********************************************************
; *** SET UP SERIAL COMMUNICATIONS WITH HOST.
; ***
;       UZI0, Port D
;
;  External driver/receiver is
;  Sipex SP3222E with output-enable and Shutdown
;  -- by disabling this chip in software, UZI0 can
;     also be used as an SPI port.
;
;Pin 1, signal COM1-ENI:  eZ80 port line Px4; (ACTIVE LOW)
;-- line is tied high through 10K
;-- configure port line as OUTPUT, OPEN DRAIN (DDR=0, ALT1=1)
;(receiver is disabled unless Active.)
;
;Pin 18, signal COM1SHDN: eZ80 port line Px5; (ACTIVE HIGH)
;-- line is tied low through 68K
;-- configure port line as OUTPUT, OPEN SOURCE (DDR=1, ALT1=1)
;( device disabled and charge pump is shut down unless Active.)
;
; activate:   xxHLxxxx  PortData & 0xCF | 0x20
; deactivate: xxLHxxxx  PortData & 0xCF | 0x10

;********************
; Define UART0 as 38400, 8 bits, no parity, 2 stop bits
;
BAUDRATE EQU 57600;38400
DIVISOR EQU ((CPUFREQ / BAUDRATE) / 16)

    LD      A,023h          ;set bits d1,d0 alternate function
    OUT0    (PD_DDR),A      ;and bit d5 for open-source I/O

    LD      A,03h           ;bits D1,D0 to Alternate function
    OUT0    (PD_ALT2),A     ;

    LD      A,10h          ;RS232 buffer/drivers disabled
    OUT0    (PD_DR),A      ;

    LD      A,30h           ; d4=open drain; d5=open source.
    OUT0    (PD_ALT1),A     ;

; ********************************************************
; ** ( initialize registers, and jump to COLD )

        LD  HL,ENDOFMEM
        dec h        ; EM-100h
        ld sp,hl     ;      = top of param stack
        inc h        ; EM
        push hl
        pop ix       ;      = top of return stack
        dec h        ; EM-200h
        dec h
        push hl
        pop iy       ;      = bottom of user area
        ld de,1      ; do reset if COLD returns
        LD  A,0Dh
        LD  (InpBuffer),A
        jp COLD      ; enter top-level Forth word

;Z INIT-UART  --    Initialize UART and activate driver/receiver
; note that both COLD and QUIT call this procedure.
;   user program can also call it if needed.
    head INIT_UART,{"INIT-UART"},docode

    LD      A,10h
    OUT0    (PD_DR),A

; ----------------------------------------------
  LD      A,01h
  OUT0    (UZI0_CTL),A    ;UZI mode = UART (01h)
  ;eZ80180 Only -- No such register in 'F92!!
; ----------------------------------------------

    LD      A,80h
    OUT0    (UART0_LCTL),A  ;enable access to BRG
    LD      A,LOW(DIVISOR)  ;LSB of BRG divisor
    OUT0    (BRG0_DLR_L),A   ;load low byte of BRG
    LD      A,HIGH(DIVISOR)  ;MSB of BRG divisor
    OUT0    (BRG0_DLR_H),A   ;load high byte of BRG
    LD      A,00h           ;
    OUT0    (UART0_LCTL),A  ;lock the BRG

    LD      A,41h           ;
    OUT0    (UART0_FCTL),A  ;enable 4-byte FIFO

    LD      A,07h
    OUT0    (UART0_LCTL),A  ;select 8bits, no parity, 1 stop

    next

; RS-232 TERMINAL I/O  =============================

UART_THRE   EQU 20h     ;bit mask for okay to Tx
UART_DR     EQU 01h     ;bit mask for Rx'd char ready

;
; transmit byte in reg C
;
TXMCHR:
$$  IN0     A,(UART0_LSR)   ; check for char waiting..
    AND     UART_THRE
    JR      Z,$B
    LD      A,C
    OUT0    (UART0_THR),A
    RET
;
; check if character ready to receive
;
RCXRDY:
        IN0     A,(UART0_LSR)   ; check for char waiting..
        AND     UART_DR
        RET

;
; receive byte into reg A
;
RCXCHR:
$$  IN0     A,(UART0_LSR)   ; check for char waiting..
    AND     UART_DR
    JR      Z,$B
    IN0     A,(UART0_RBR)
    RET

; INTERPRETER LOGIC =============================
; See also "defining words" at end of this file

;C EXIT     --      exit a colon definition
    head EXIT,{"EXIT"},docode
        ld e,(ix+0)    ; pop old IP from ret stk
        inc ix
        ld d,(ix+0)
        inc ix
        next

;Z lit      -- x    fetch inline literal to stack
; This is the primitive compiled by LITERAL.
    head LIT,{"lit"},docode
        push bc        ; push old TOS
        ld a,(de)      ; fetch cell at IP to TOS,
        ld c,a         ;        advancing IP
        inc de
        ld a,(de)
        ld b,a
        inc de
        next

;C EXECUTE   i*x xt -- j*x   execute Forth word
;C                           at 'xt'
    head EXECUTE,{"EXECUTE"},docode
        ld h,b          ; address of word -> HL
        ld l,c
        pop bc          ; get new TOS
        jp (hl)         ; go do Forth word

; DEFINING WORDS ================================

; ENTER, a.k.a. DOCOLON, entered by CALL ENTER
; to enter a new high-level thread (colon def'n.)
; (internal code fragment, not a Forth word)
; N.B.: DOCOLON must be defined before any
; appearance of 'docolon' in a 'word' macro!
docolon:               ; (alternate name)
enter:  dec ix         ; push old IP on ret stack
        ld (ix+0),d
        dec ix
        ld (ix+0),e
        pop hl         ; param field adrs -> IP
        nexthl         ; use the faster 'nexthl'

;C VARIABLE   --      define a Forth variable
;   CREATE 1 CELLS ALLOT ;
; Action of RAM variable is identical to CREATE,
; so we don't need a DOES> clause to change it.
    head VARIABLE,{"VARIABLE"},docolon
        DW CREATE,LIT,1,CELLS,ALLOT,EXIT
; DOVAR, code action of VARIABLE, entered by CALL
; DOCREATE, code action of newly created words
docreate:
dovar:  ; -- a-addr
        pop hl     ; parameter field address
        push bc    ; push old TOS
        ld b,h     ; pfa = variable's adrs -> TOS
        ld c,l
        next

;C CONSTANT   n --      define a Forth constant
;   CREATE , DOES> (machine code fragment)
    head CONSTANT,{"CONSTANT"},docolon
        DW CREATE,COMMA,XDOES
; DOCON, code action of CONSTANT,
; entered by CALL DOCON
docon:  ; -- x
        pop hl     ; parameter field address
        push bc    ; push old TOS
        ld c,(hl)  ; fetch contents of parameter
        inc hl     ;    field -> TOS
        ld b,(hl)
        next

;Z USER     n --        define user variable 'n'
;   CREATE , DOES> (machine code fragment)
    head USER,{"USER"},docolon
        DW CREATE,COMMA,XDOES
; DOUSER, code action of USER,
; entered by CALL DOUSER
douser:  ; -- a-addr
        pop hl     ; parameter field address
        push bc    ; push old TOS
        ld c,(hl)  ; fetch contents of parameter
        inc hl     ;    field
        ld b,(hl)
        push iy    ; copy user base address to HL
        pop hl
        add hl,bc  ;    and add offset
        ld b,h     ; put result in TOS
        ld c,l
        next

; DODOES, code action of DOES> clause
; entered by       CALL fragment
;                  parameter field
;                       ...
;        fragment: CALL DODOES
;                  high-level thread
; Enters high-level thread with address of
; parameter field on top of stack.
; (internal code fragment, not a Forth word)
dodoes: ; -- a-addr
        dec ix         ; push old IP on ret stk
        ld (ix+0),d
        dec ix
        ld (ix+0),e
        pop de         ; adrs of new thread -> IP
        pop hl         ; adrs of parameter field
        push bc        ; push old TOS onto stack
        ld b,h         ; pfa -> new TOS
        ld c,l
        next


;C EMIT     c --    output character to console

    head EMIT,{"EMIT"},docode
        CALL    TXMCHR
        POP     BC
        next

;C KEY?     -- f    return true if char waiting
    head QUERYKEY,{"KEY?"},docode
        PUSH BC     ;PREPARE FOR NEW TOS
        CALL RCXRDY
        LD C,A      ; RESULT IN TOS
        LD B,0
        next

;C KEY      -- c    get character from keyboard
    head KEY,{"KEY"},docode
        PUSH BC     ;PREPARE FOR NEW TOS
        CALL RCXCHR
        LD C,A      ; RESULT IN TOS
        LD B,0
        next


;X BYE     i*x --    return to CP/M
    head BYE,{"bye"},docode

        HALT
        jp 0

; STACK OPERATIONS ==============================

;C DUP      x -- x x      duplicate top of stack
    head DUP,{"DUP"},docode
pushtos: push bc
        next

;C ?DUP     x -- 0 | x x    DUP if nonzero
    head QDUP,{"?DUP"},docode
        ld a,b
        or c
        jr nz,pushtos
        next

;C DROP     x --          drop top of stack
    head DROP,{"DROP"},docode
poptos: pop bc
        next

;C SWAP     x1 x2 -- x2 x1    swap top two items
    head SWOP,{"SWAP"},docode
        pop hl
        push bc
        ld b,h
        ld c,l
        next

;C OVER    x1 x2 -- x1 x2 x1   per stack diagram
    head OVER,{"OVER"},docode
        pop hl
        push hl
        push bc
        ld b,h
        ld c,l
        next

;C ROT    x1 x2 x3 -- x2 x3 x1  per stack diagram
    head ROT,{"ROT"},docode
        ; x3 is in TOS
        pop hl          ; x2
        ex (sp),hl      ; x2 on stack, x1 in hl
        push bc
        ld b,h
        ld c,l
        next

;X NIP    x1 x2 -- x2           per stack diagram
    head NIP,{"NIP"},docolon
        DW SWOP,DROP,EXIT

;X TUCK   x1 x2 -- x2 x1 x2     per stack diagram
    head TUCK,{"TUCK"},docolon
        DW SWOP,OVER,EXIT

;C >R    x --   R: -- x   push to return stack
    head TOR,{">R"},docode
        dec ix          ; push TOS onto rtn stk
        ld (ix+0),b
        dec ix
        ld (ix+0),c
        pop bc          ; pop new TOS
        next

;C R>    -- x    R: x --   pop from return stack
    head RFROM,{"R>"},docode
        push bc         ; push old TOS
        ld c,(ix+0)     ; pop top rtn stk item
        inc ix          ;       to TOS
        ld b,(ix+0)
        inc ix
        next

;C R@    -- x     R: x -- x   fetch from rtn stk
    head RFETCH,{"R@"},docode
        push bc         ; push old TOS
        ld c,(ix+0)     ; fetch top rtn stk item
        ld b,(ix+1)     ;       to TOS
        next

;Z SP@  -- a-addr       get data stack pointer
    head SPFETCH,{"SP@"},docode
        push bc
        ld hl,0
        add hl,sp
        ld b,h
        ld c,l
        next

;Z SP!  a-addr --       set data stack pointer
    head SPSTORE,{"SP!"},docode
        ld h,b
        ld l,c
        ld sp,hl
        pop bc          ; get new TOS
        next

;Z RP@  -- a-addr       get return stack pointer
    head RPFETCH,{"RP@"},docode
        push bc
        push ix
        pop bc
        next

;Z RP!  a-addr --       set return stack pointer
    head RPSTORE,{"RP!"},docode
        push bc
        pop ix
        pop bc
        next

; MEMORY AND I/O OPERATIONS =====================

;C !        x a-addr --   store cell in memory
    head STORE,{'!'},docode
        ld h,b          ; address in hl
        ld l,c
        pop bc          ; data in bc
        ld (hl),c
        inc hl
        ld (hl),b
        pop bc          ; pop new TOS
        next

;C C!      char c-addr --    store char in memory
    head CSTORE,{"C!"},docode
        ld h,b          ; address in hl
        ld l,c
        pop bc          ; data in bc
        ld (hl),c
        pop bc          ; pop new TOS
        next

;C @       a-addr -- x   fetch cell from memory
    head FETCH,{'@'},docode
        ld h,b          ; address in hl
        ld l,c
        ld c,(hl)
        inc hl
        ld b,(hl)
        next

;C C@     c-addr -- char   fetch char from memory
    head CFETCH,{"C@"},docode
        ld a,(bc)
        ld c,a
        ld b,0
        next

; ---------- PORT I/O Commands

 ;C  P!  ( n p -- ) write byte n to i/o port p

    head PSTORE,{"P!"},docode
        POP     HL
        OUT     (C),L
        POP     BC      ; get new TOS
        next

;C P@   ( p -- n )  read byte n from i/o port p

    head PFETCH,{"P@"},docode
        IN C,(C)        ;simple stuff...
        LD B,0
        next

            COMMENT %
;C  P0!  ( n p -- ) write byte n to Z18x internal i/o port p

    head P0STORE,{"P0!"},docode
        POP     HL
        LD  A,C
        LD ($F+2),A     ;plug into OUT0 opcode
$$      OUT0    (7Fh),L
        POP     BC      ; get new TOS
        next

;C P0@   ( p -- n )  read byte n from Z18x internal i/o port p

    head P0FETCH,{"P0@"},docode
        LD  A,C
        LD ($F+2),A     ;plug into IN0 opcode
$$      IN0     C,(7Fh)
        LD      B,0
        next


;Z PC!     char c-addr --    output char to port
    head PCSTORE,{"PC!"},docode
        pop hl          ; char in L
        out (c),l       ; to port (BC)
        pop bc          ; pop new TOS
        next

;Z PC@     c-addr -- char   input char from port
    head PCFETCH,{"PC@"},docode
        in c,(c)        ; read port (BC) to C
        ld b,0
        next
  % ENDCOMMENT

; ARITHMETIC AND LOGICAL OPERATIONS =============

;C +       n1/u1 n2/u2 -- n3/u3     add n1+n2
    head PLUS,{'+'},docode
        pop hl
        add hl,bc
        ld b,h
        ld c,l
        next

;X M+       d n -- d         add single to double
    head MPLUS,{"M+"},docode
        ex de,hl
        pop de          ; hi cell
        ex (sp),hl      ; lo cell, save IP
        add hl,bc
        ld b,d          ; hi result in BC (TOS)
        ld c,e
        jr nc,mplus1
        inc bc
mplus1: pop de          ; restore saved IP
        push hl         ; push lo result
        next

;C -      n1/u1 n2/u2 -- n3/u3    subtract n1-n2
    head MINUS,{'-'},docode
        pop hl
        or a
        sbc hl,bc
        ld b,h
        ld c,l
        next

;C AND    x1 x2 -- x3            logical AND
    head AND,{"AND"},docode
        pop hl
        ld a,b
        and h
        ld b,a
        ld a,c
        and l
        ld c,a
        next

;C OR     x1 x2 -- x3           logical OR
    head OR,{"OR"},docode
        pop hl
        ld a,b
        or h
        ld b,a
        ld a,c
        or l
        ld c,a
        next

;C XOR    x1 x2 -- x3            logical XOR
    head XOR,{"XOR"},docode
        pop hl
        ld a,b
        xor h
        ld b,a
        ld a,c
        xor l
        ld c,a
        next

;C INVERT   x1 -- x2            bitwise inversion
    head INVERT,{"INVERT"},docode
        ld a,b
        cpl
        ld b,a
        ld a,c
        cpl
        ld c,a
        next

;C NEGATE   x1 -- x2            two's complement
    head NEGATE,{"NEGATE"},docode
        ld a,b
        cpl
        ld b,a
        ld a,c
        cpl
        ld c,a
        inc bc
        next

;C 1+      n1/u1 -- n2/u2       add 1 to TOS
    head ONEPLUS,{"1+"},docode
        inc bc
        next

;C 1-      n1/u1 -- n2/u2     subtract 1 from TOS
    head ONEMINUS,{"1-"},docode
        dec bc
        next

;Z ><      x1 -- x2         swap bytes (not ANSI)
    head SWAPBYTES,{"><"},docode
        ld a,b
        ld b,c
        ld c,a
        next

;C 2*      x1 -- x2         arithmetic left shift
    head TWOSTAR,{"2*"},docode
        sla c
        rl b
        next

;C 2/      x1 -- x2        arithmetic right shift
    head TWOSLASH,{"2/"},docode
        sra b
        rr c
        next

;C LSHIFT  x1 u -- x2    logical L shift u places
    head LSHIFT,{"LSHIFT"},docode
        ld b,c        ; b = loop counter
        pop hl        ;   NB: hi 8 bits ignored!
        inc b         ; test for counter=0 case
        jr lsh2
lsh1:   add hl,hl     ; left shift HL, n times
lsh2:   djnz lsh1
        ld b,h        ; result is new TOS
        ld c,l
        next

;C RSHIFT  x1 u -- x2    logical R shift u places
    head RSHIFT,{"RSHIFT"},docode
        ld b,c        ; b = loop counter
        pop hl        ;   NB: hi 8 bits ignored!
        inc b         ; test for counter=0 case
        jr rsh2
rsh1:   srl h         ; right shift HL, n times
        rr l
rsh2:   djnz rsh1
        ld b,h        ; result is new TOS
        ld c,l
        next

;C +!     n/u a-addr --       add cell to memory
    head PLUSSTORE,{"+!"},docode
        pop hl
        ld a,(bc)       ; low byte
        add a,l
        ld (bc),a
        inc bc
        ld a,(bc)       ; high byte
        adc a,h
        ld (bc),a
        pop bc          ; pop new TOS
        next

; COMPARISON OPERATIONS =========================

;C 0=     n/u -- flag    return true if TOS=0
    head ZEROEQUAL,{"0="},docode
        ld a,b
        or c            ; result=0 if bc was 0
        sub 1           ; cy set   if bc was 0
        sbc a,a         ; propagate cy through A
        ld b,a          ; put 0000 or FFFF in TOS
        ld c,a
        next

;C 0<     n -- flag      true if TOS negative
    head ZEROLESS,{"0<"},docode
        sla b           ; sign bit -> cy flag
        sbc a,a         ; propagate cy through A
        ld b,a          ; put 0000 or FFFF in TOS
        ld c,a
        next

;C =      x1 x2 -- flag         test x1=x2
    head EQUAL,{'='},docode
        pop hl
        or a
        sbc hl,bc       ; x1-x2 in HL, SZVC valid
        jr z,tostrue
tosfalse: ld bc,0
        next

;X <>     x1 x2 -- flag    test not eq (not ANSI)
    head NOTEQUAL,{"<>"},docolon
        DW EQUAL,ZEROEQUAL,EXIT

;C <      n1 n2 -- flag        test n1<n2, signed
    head LESS,{'<'},docode
        pop hl
        or a
        sbc hl,bc       ; n1-n2 in HL, SZVC valid
; if result negative & not OV, n1<n2
; neg. & OV => n1 +ve, n2 -ve, rslt -ve, so n1>n2
; if result positive & not OV, n1>=n2
; pos. & OV => n1 -ve, n2 +ve, rslt +ve, so n1<n2
; thus OV reverses the sense of the sign bit
        jp pe,revsense  ; if OV, use rev. sense
        jp p,tosfalse   ;   if +ve, result false
tostrue: ld bc,0ffffh   ;   if -ve, result true
        next
revsense: jp m,tosfalse ; OV: if -ve, reslt false
        jr tostrue      ;     if +ve, result true

;C >     n1 n2 -- flag         test n1>n2, signed
    head GREATER,{'>'},docolon
        DW SWOP,LESS,EXIT

;C U<    u1 u2 -- flag       test u1<n2, unsigned
    head ULESS,{"U<"},docode
        pop hl
        or a
        sbc hl,bc       ; u1-u2 in HL, SZVC valid
        sbc a,a         ; propagate cy through A
        ld b,a          ; put 0000 or FFFF in TOS
        ld c,a
        next

;X U>    u1 u2 -- flag     u1>u2 unsgd (not ANSI)
    head UGREATER,{"U>"},docolon
        DW SWOP,ULESS,EXIT

; LOOP AND BRANCH OPERATIONS ====================

;Z branch   --                  branch always
    head BRANCH,{"branch"},docode
dobranch: ld a,(de)     ; get inline value => IP
        ld l,a
        inc de
        ld a,(de)
        ld h,a
        nexthl

;Z ?branch   x --              branch if TOS zero
    head QBRANCH,{"?branch"},docode
        ld a,b
        or c            ; test old TOS
        pop bc          ; pop new TOS
        jr z,dobranch   ; if old TOS=0, branch
        inc de          ; else skip inline value
        inc de
        next

;Z (do)    n1|u1 n2|u2 --  R: -- sys1 sys2
;Z                          run-time code for DO
; '83 and ANSI standard loops terminate when the
; boundary of limit-1 and limit is crossed, in
; either direction.  This can be conveniently
; implemented by making the limit 8000h, so that
; arithmetic overflow logic can detect crossing.
; I learned this trick from Laxen & Perry F83.
; fudge factor = 8000h-limit, to be added to
; the start value.
    head XDO,{"(do)"},docode
        ex de,hl
        ex (sp),hl   ; IP on stack, limit in HL
        ex de,hl
        ld hl,8000h
        or a
        sbc hl,de    ; 8000-limit in HL
        dec ix       ; push this fudge factor
        ld (ix+0),h  ;    onto return stack
        dec ix       ;    for later use by 'I'
        ld (ix+0),l
        add hl,bc    ; add fudge to start value
        dec ix       ; push adjusted start value
        ld (ix+0),h  ;    onto return stack
        dec ix       ;    as the loop index.
        ld (ix+0),l
        pop de       ; restore the saved IP
        pop bc       ; pop new TOS
        next

;Z (loop)   R: sys1 sys2 --  | sys1 sys2
;Z                        run-time code for LOOP
; Add 1 to the loop index.  If loop terminates,
; clean up the return stack and skip the branch.
; Else take the inline branch.  Note that LOOP
; terminates when index=8000h.
    head XLOOP,{"(loop)"},docode
        exx
        ld bc,1
looptst: ld l,(ix+0)  ; get the loop index
        ld h,(ix+1)
        or a
        adc hl,bc    ; increment w/overflow test
        jp pe,loopterm  ; overflow=loop done
        ; continue the loop
        ld (ix+0),l  ; save the updated index
        ld (ix+1),h
        exx
        jr dobranch  ; take the inline branch
loopterm: ; terminate the loop
        ld bc,4      ; discard the loop info
        add ix,bc
        exx
        inc de       ; skip the inline branch
        inc de
        next

;Z (+loop)   n --   R: sys1 sys2 --  | sys1 sys2
;Z                        run-time code for +LOOP
; Add n to the loop index.  If loop terminates,
; clean up the return stack and skip the branch.
; Else take the inline branch.
    head XPLUSLOOP,{"(+loop)"},docode
        pop hl      ; this will be the new TOS
        push bc
        ld b,h
        ld c,l
        exx
        pop bc      ; old TOS = loop increment
        jr looptst

;C I        -- n   R: sys1 sys2 -- sys1 sys2
;C                  get the innermost loop index
    head II,{'I'},docode
        push bc     ; push old TOS
        ld l,(ix+0) ; get current loop index
        ld h,(ix+1)
        ld c,(ix+2) ; get fudge factor
        ld b,(ix+3)
        or a
        sbc hl,bc   ; subtract fudge factor,
        ld b,h      ;   returning true index
        ld c,l
        next

;C J        -- n   R: 4*sys -- 4*sys
;C                  get the second loop index
    head JJ,{'J'},docode
        push bc     ; push old TOS
        ld l,(ix+4) ; get current loop index
        ld h,(ix+5)
        ld c,(ix+6) ; get fudge factor
        ld b,(ix+7)
        or a
        sbc hl,bc   ; subtract fudge factor,
        ld b,h      ;   returning true index
        ld c,l
        next

;C UNLOOP   --   R: sys1 sys2 --  drop loop parms
    head UNLOOP,{"UNLOOP"},docode
        inc ix
        inc ix
        inc ix
        inc ix
        next

; MULTIPLY AND DIVIDE ===========================

;C UM*     u1 u2 -- ud   unsigned 16x16->32 mult.
    head UMSTAR,{"UM*"},docode
        push bc
        exx
        pop bc      ; u2 in BC
        pop de      ; u1 in DE
        ld hl,0     ; result will be in HLDE
        ld a,17     ; loop counter
        or a        ; clear cy
umloop: rr h
        rr l
        rr d
        rr e
        jr nc,noadd
        add hl,bc
noadd:  dec a
        jr nz,umloop
        push de     ; lo result
        push hl     ; hi result
        exx
        pop bc      ; put TOS back in BC
        next

;C UM/MOD   ud u1 -- u2 u3   unsigned 32/16->16
    head UMSLASHMOD,{"UM/MOD"},docode
        push bc
        exx
        pop bc      ; BC = divisor
        pop hl      ; HLDE = dividend
        pop de
        ld a,16     ; loop counter
        sla e
        rl d        ; hi bit DE -> carry
udloop: adc hl,hl   ; rot left w/ carry
        jr nc,udiv3
        ; case 1: 17 bit, cy:HL = 1xxxx
        or a        ; we know we can subtract
        sbc hl,bc
        or a        ; clear cy to indicate sub ok
        jr udiv4
        ; case 2: 16 bit, cy:HL = 0xxxx
udiv3:  sbc hl,bc   ; try the subtract
        jr nc,udiv4 ; if no cy, subtract ok
        add hl,bc   ; else cancel the subtract
        scf         ;   and set cy to indicate
udiv4:  rl e        ; rotate result bit into DE,
        rl d        ; and next bit of DE into cy
        dec a
        jr nz,udloop
        ; now have complemented quotient in DE,
        ; and remainder in HL
        ld a,d
        cpl
        ld b,a
        ld a,e
        cpl
        ld c,a
        push hl     ; push remainder
        push bc
        exx
        pop bc      ; quotient remains in TOS
        next

; BLOCK AND STRING OPERATIONS ===================

;C FILL   c-addr u char --  fill memory with char
    head FILL,{"FILL"},docode
        ld a,c          ; character in a
        exx             ; use alt. register set
        pop bc          ; count in bc
        pop de          ; address in de
        or a            ; clear carry flag
        ld hl,0ffffh
        adc hl,bc       ; test for count=0 or 1
        jr nc,filldone  ;   no cy: count=0, skip
        ld (de),a       ; fill first byte
        jr z,filldone   ;   zero, count=1, done
        dec bc          ; else adjust count,
        ld h,d          ;   let hl = start adrs,
        ld l,e
        inc de          ;   let de = start adrs+1
        ldir            ;   copy (hl)->(de)
filldone: exx           ; back to main reg set
        pop bc          ; pop new TOS
        next

;X CMOVE   c-addr1 c-addr2 u --  move from bottom
; as defined in the ANSI optional String word set
; On byte machines, CMOVE and CMOVE> are logical
; factors of MOVE.  They are easy to implement on
; CPUs which have a block-move instruction.
    head CMOVE,{"CMOVE"},docode
        push bc
        exx
        pop bc      ; count
        pop de      ; destination adrs
        pop hl      ; source adrs
        ld a,b      ; test for count=0
        or c
        jr z,cmovedone
        ldir        ; move from bottom to top
cmovedone: exx
        pop bc      ; pop new TOS
        next

;X CMOVE>  c-addr1 c-addr2 u --  move from top
; as defined in the ANSI optional String word set
    head CMOVEUP,{"CMOVE>"},docode
        push bc
        exx
        pop bc      ; count
        pop hl      ; destination adrs
        pop de      ; source adrs
        ld a,b      ; test for count=0
        or c
        jr z,umovedone
        add hl,bc   ; last byte in destination
        dec hl
        ex de,hl
        add hl,bc   ; last byte in source
        dec hl
        lddr        ; move from top to bottom
umovedone: exx
        pop bc      ; pop new TOS
        next

;Z SKIP   c-addr u c -- c-addr' u'
;Z                          skip matching chars
; Although SKIP, SCAN, and S= are perhaps not the
; ideal factors of WORD and FIND, they closely
; follow the string operations available on many
; CPUs, and so are easy to implement and fast.
    head SKIP,{"SKIP"},docode
        ld a,c      ; skip character
        exx
        pop bc      ; count
        pop hl      ; address
        ld e,a      ; test for count=0
        ld a,b
        or c
        jr z,skipdone
        ld a,e
skiploop:
        EX  AF,AF'
        EX  AF,AF'
        CPI
        jr nz,skipmis   ; char mismatch: exit
        jp pe,skiploop  ; count not exhausted
        jr skipdone     ; count 0, no mismatch
skipmis: inc bc         ; mismatch!  undo last to
        dec hl          ;  point at mismatch char
skipdone: push hl   ; updated address
        push bc     ; updated count
        exx
        pop bc      ; TOS in bc
        next

;Z SCAN    c-addr u c -- c-addr' u'
;Z                      find matching char
    head SCAN,{"SCAN"},docode
        ld a,c      ; scan character
        exx
        pop bc      ; count
        pop hl      ; address
        ld e,a      ; test for count=0
        ld a,b
        or c
        jr z,scandone
        ld a,e
        CPIR        ; scan 'til match or count=0
        jr nz,scandone  ; no match, BC & HL ok
        inc bc          ; match!  undo last to
        dec hl          ;   point at match char
scandone: push hl   ; updated address
        push bc     ; updated count
        exx
        pop bc      ; TOS in bc
        next

;Z S=    c-addr1 c-addr2 u -- n   string compare
;Z             n<0: s1<s2, n=0: s1=s2, n>0: s1>s2
    head SEQUAL,{"S="},docode
        push bc
        exx
        pop bc      ; count
        pop hl      ; addr2
        pop de      ; addr1
        ld a,b      ; test for count=0
        or c
        jr z,smatch     ; by definition, match!
sloop:
        ld a,(de)
        inc de
        CPI
        jr nz,sdiff     ; char mismatch: exit
        jp pe,sloop     ; count not exhausted
smatch: ; count exhausted & no mismatch found
        exx
        ld bc,0         ; bc=0000  (s1=s2)
        jr snext
sdiff:  ; mismatch!  undo last 'cpi' increment
        dec hl          ; point at mismatch char
        cp (hl)         ; set cy if char1 < char2
        sbc a,a         ; propagate cy thru A
        exx
        ld b,a          ; bc=FFFF if cy (s1<s2)
        or 1            ; bc=0001 if ncy (s1>s2)
        ld c,a
snext:  next


 COMMENT %
;Z WRSEC   src dst -- src+256 dst+256
;Z   write 256-byte sector to AT29C040A flash
;    returned values are n
    head WRSEC,{"WRSEC"},docode
        LD  A,00h
        POP HL      ;set source address (src)
        PUSH DE     ; save IP (Instruction Pointer)
        LD  E,C     ;TOS into DE (dst)
        LD  D,B
        LD  BC,256  ;set count
        LD  A,0AAh ; .....
        LD  (5555h),A  ;write-enable algorithm
        LD  A,55h
        LD  (2AAAh),A
        LD  A,0A0h
        LD  (5555h),A
        LDIR        ;move 256 bytes to Flash ROM
        DEC HL      ;point back to last byte programmed
        DEC DE
 $$     LD  A,(DE)  ;compare last written byte,
        XOR (HL)    ;use only bit d7... it
        AND 80h     ;will toggle until end of
        JR  NZ,$B   ;  internal write
        INC HL      ; bump forward again
        INC DE      ;bump forward again
        LD  C,E     ;DE (dst) to TOS
        LD  B,D
        POP DE     ; restore IP (Instruction Pointer)
        PUSH HL    ; set 2nd stack item (src)
        next
 % ENDCOMMENT


        .INCLUDE "CAML192d.S"   ; CPU Dependencies
        .INCLUDE "CAML192h.S"   ; High Level words

lastword EQU link       ; nfa of last word in dict.

    ;SEGMENT Onchip
   	;ORG 0E000h

InpBuffer DS 128 ;BLKB 128

enddict: ; EQU $       ; user's code starts here

        END     ;ENTRY
